package es.rediris.papi;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.security.KeyPair;
import java.security.Provider;
import java.util.Properties;
import javax.servlet.ServletException;
import org.bouncycastle.crypto.engines.AESEngine;
import org.bouncycastle.crypto.engines.RSAEngine;
import org.bouncycastle.crypto.params.KeyParameter;
import org.bouncycastle.crypto.params.RSAKeyParameters;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.jce.provider.JCERSAPrivateKey;
import org.bouncycastle.jce.provider.JCERSAPublicKey;
import org.bouncycastle.openssl.PEMReader;
import org.bouncycastle.util.encoders.Base64;
import org.bouncycastle.util.encoders.Hex;
import org.bouncycastle.crypto.digests.SHA1Digest;

/** Cryptographic class for PAPIFilter. Includes AES and RSA encoding/decoding
  * @author ajay
 */
public class Ciphers {
    
    private BouncyCastleProvider bcp;
    
    private RSAEngine rsa;
    private AESEngine aes;
    private SHA1Digest sha1;
    
    private RSAKeyParameters rsaPublicKey, rsaPrivateKey;
    
    private KeyParameter hkey, lkey;
    
    private Properties papiConf;
    
    
    /** Creates a new instance of Ciphers . Uses default falues set in the PAPI configuration file. Initializes AES hkey and lkey files.
     * @param papiConf Configuration values
     * @throws ServletException if there's been any exception while creating the engines or reading the key files
     */
    public Ciphers(Properties papiConf) throws ServletException {
        
        try {
         
            // register the security provider
            this.papiConf = papiConf;    
            bcp = new BouncyCastleProvider();
            java.security.Security.addProvider((Provider)bcp);

            // create the engines
            rsa = new RSAEngine();
            aes = new AESEngine();
            sha1 = new SHA1Digest();
            
            // read the hkey and lkey files for AES encoding/decoding
            BufferedReader aesFile = new BufferedReader(new FileReader((String)papiConf.get("PoAkeys.hkeyName")));
            hkey = new KeyParameter(aesFile.readLine().getBytes());
            aesFile.close();
            aesFile = new BufferedReader(new FileReader((String)papiConf.get("PoAkeys.lkeyName")));
            lkey = new KeyParameter(aesFile.readLine().getBytes());
            aesFile.close();
        }
        catch (IOException e) {throw new ServletException ("Exception in Cipher constructor: " + e.getMessage());}

    }
    
    /** Initializes the RSA engine, used with GPoA data and AS assertions
     * @param ASname name of the signer
     * @throws ServletException
     */
    public void init(String ASname) throws ServletException {
        
        try {
                
            // get the file path
            String pubkey = (String)papiConf.get(ASname + ".pubkeyName");
            
            // create public key parameters
            PEMReader keyFile = new PEMReader (new BufferedReader( new FileReader(pubkey)));
            JCERSAPublicKey publicPass = (JCERSAPublicKey)keyFile.readObject();
            rsaPublicKey = new RSAKeyParameters(false, publicPass.getModulus(), publicPass.getPublicExponent());
            keyFile.close();

            if (papiConf.containsKey(ASname + ".privkeyName")) {
                // create private key parameters
                String privkey = (String)papiConf.get(ASname + ".privkeyName");
                keyFile = new PEMReader (new BufferedReader( new FileReader(privkey)));
                KeyPair privatePair = (KeyPair)keyFile.readObject();
                JCERSAPrivateKey privatePass = (JCERSAPrivateKey)privatePair.getPrivate();
                rsaPrivateKey = new RSAKeyParameters(true, privatePass.getModulus(), privatePass.getPrivateExponent());
                keyFile.close();
            }
            
            sha1.reset();
        } 
        catch (IOException e) {throw new ServletException ("Exception in Cipher init: " + e.getMessage());}
        
    }

    /** RSA encoding/decoding
     * @param encoding true if we wish to encode data
     * @param data data to encode/decode
     * @return ciphered data
     * @throws ServletException
     */
    public String RSA_Encode(boolean encoding, String data) throws ServletException {
    
        int blockSize = 0;
        int blocks = 0;
        byte[] rawData = null;
        String text = "";
        
        try {
        
                // if encoding is selected, initialize the engine with the private key
                if (encoding) {
                    rsa.init(encoding, rsaPrivateKey);
                    blockSize = rsa.getOutputBlockSize();
                    rawData = data.getBytes();
                }

                // if decoding is selected, initialize the engine with the public key
                else {
                    rsa.init(encoding, rsaPublicKey);
                    blockSize = rsa.getInputBlockSize();
                    rawData = Base64.decode(data);
                }

                //decompose the incoming data in blocks
                int rawDataSize = rawData.length;
                blocks = rawDataSize / blockSize;
                if (rawDataSize % blockSize != 0) {
                    blocks++;
                    int tmpRawDataSize = blocks*blockSize;
                    byte[] tmpRawData = new byte[tmpRawDataSize];

                    for (int i=0; i<rawDataSize; i++) {
                        tmpRawData[i] = rawData[i];
                    }
                    for (int i=rawDataSize; i< tmpRawDataSize; i++) {
                        tmpRawData[i] = -1;
                    }

                    rawData = tmpRawData;
                    rawDataSize = tmpRawDataSize;
                }

                // process each block
                byte[][] outputMatrix = new byte[blocks][blockSize]; //= new byte[blockSize];
                int blockCount = 0;
                int count = 0;
                while (count < rawDataSize) {
                    outputMatrix[blockCount] = rsa.processBlock(rawData, count, blockSize);
                    count += blockSize;
                    blockCount++;
                }

                // rebuild data
                byte[] output = new byte[blocks*blockSize];
                for (int i=0; i<blocks; i++) {
                    for (int j=0; j<blockSize; j++) {
                        if (j < outputMatrix[i].length)
                            output[i*blocks + j] = outputMatrix[i][j];
                        else
                            output[i*blocks + j] = -1;
                    }
                }

                if (encoding)
                    text = new String(Base64.encode(output));
                else {
                    String fullText = new String(output);
                    text = removeRSAPaddingChars(fullText);
                }
        }
        catch (Exception e) {throw new ServletException ("Exception in RSA encoding/decoding: " + e.getMessage ());}
        
        return text;        
        
    }
    
    /** Remove padding characters to get clean data
     * @param data Information to clean
     * @return clean data
     */
    private String removeRSAPaddingChars(String data) {
        byte[] paddingChar1 = new byte[1];
        byte[] paddingChar2 = new byte[1];
        byte[] paddingChar3 = new byte[1];
        
        // define the 3 types of padding characters used
        paddingChar1[0] = -1;
        paddingChar2[0] = 0;
        paddingChar3[0] = 1;
        
        String aux1 = data.replaceAll(new String(paddingChar1), "");
        String aux2 = aux1.replaceAll(new String(paddingChar2), "");
        String output = aux2.replaceAll(new String(paddingChar3), "");
        return output;
    }
    
    /** AES encoding/decoding
     * @param encode True if we wish to encode
     * @param isHcook True if the data to process belongs to a Hcook
     * @param data Data to process
     * @return Encoded/decoded data
     * @throws ServletException
     */
    public String AES_Encode(boolean encode, boolean isHcook, String data) throws ServletException  {
        
        String codedOutput = "";
        aes.reset();
        
        try {
         
            //initialize with the corresponding key
            if (isHcook)
                aes.init(encode, hkey);
            else
                aes.init(encode, lkey);
            
            int count = 0;
            int blockSize = aes.getBlockSize();
            
            byte[] rawData = null;
            if (encode) {
                rawData = data.getBytes();
            } else {
                byte[] aux = data.getBytes();
                rawData = Base64.decode(aux);
            }

            // decompose the data in blocks
            int iterations = rawData.length / blockSize;
            
            if (rawData.length % blockSize != 0) {
                iterations++;
                byte[] aux = new byte[blockSize*iterations];
                for (int i=0; i<rawData.length; i++) {
                    aux[i] = rawData[i];
                }
                for (int i=rawData.length; i < aux.length; i++) {
                    aux[i] = 0;
                }
                
                rawData = aux;
            }
            
            byte[] output = new byte[rawData.length];
            
            // process each block
            for (int i=0; i<iterations; i++) {
                aes.processBlock(rawData, count, output, count);
                count += blockSize;
            }
            
            
            if (encode) {
                output = Base64.encode(output);
            }
            
            codedOutput = new String(output);
            if (!encode) {
                String aux = codedOutput.trim();
                codedOutput = aux;
            }
        }
        catch (Exception e) {throw new ServletException ("Exception in AES encoding/decoding: " + e.getMessage ());}
        
        return codedOutput;
    }

    public String createDigest(String data) {
    	String digest = "";
    	byte[] strData = data.getBytes();
    	sha1.update(strData, 0, strData.length);
    	byte[] output = new byte[sha1.getDigestSize()];
    	sha1.doFinal(output, 0);
    	byte[] hexValue = Hex.encode(output);
    	digest = new String(hexValue);
    	
    	return digest;
    }
    
}
